﻿ using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace EAA
{
    public partial class Form1 : Form
    {

        PictureBox ReferenceObject;
        MathFunc selectedFunc;
        List<MathFunc> placedFuncs = new List<MathFunc>();
        Tool curTool = Tool.None;
        Output output;
        Grid grid;
        Hysteresis hysterese;
        List<PointF> areaLabelPos = new List<PointF>();
        public enum Extrema { Max, Min };
        enum Tool { Edit, AddFunc, GetArea, Hysteresis, None};
        List<EnclosedArea> embeddedAreas = new List<EnclosedArea>();
        List<double> time = new List<double>();
        List<double> b = new List<double>();
        List<double> h = new List<double>();
        bool Tool_InEditMode = false;
        PointF MousePos = new PointF(0, 0);
        bool helpMode = false;
        bool drawHysterese = false; 
        

        public Form1()
        {
            InitializeComponent();
            InitializeVars();
        }

        public void InitializeVars()
        {
            richTextBoxOutput.Font = new Font("", 10, FontStyle.Regular);
            output = new Output(richTextBoxOutput, pictureBox1);
            grid = new Grid(pictureBox1);
            hysterese = new Hysteresis(output,pictureBox1); 
            toolStripButtonEdit.Text = "Edit";
            toolStripButtonSin.Text = "Sine";
            toolStripButtonXPowEven.Text = "Exp1";
            toolStripButtonXPowOdd.Text = "Exp2";
            ReferenceObject = pictureBox1;
            toolStripButtonEdit.Tag = "Tool: Edit\nAllows you to edit a function by clicking into the blue cycle around it.";
            toolStripButtonCalcArea.Tag = "Tool: Get Area\nAfter selecting this tool click into an encased area of two functions to get its value.";
            toolStripButtonHysteresis.Tag = "Tool: Load Hysteresis\nLoad a .lab file from your Cassy Lab device to draw the recorded hysteresis loop.";
            toolStripButtonSin.Tag = "Tool: Sine\nGenerates a sine functions. Change its parameter by scrolling along the scrollbars at the top.";
            toolStripButtonXPowEven.Tag = "Tool: Parabola 1\nGenerates a x²  function. Change its parameter by scrolling along the scrollbars at the top.";
            toolStripButtonXPowOdd.Tag = "Tool: Parabola 2\nGenerates a x³  function. Change its parameter by scrolling along the scrollbars at the top.";
            toolStripButtonCircle.Tag = "Tool: Circle\nGenerates a semicircle. Change its radius by scrolling along the scrollbars at the top.";
            toolStripButtonLinear.Tag = "Tool: Linear\nGenerates a linear  function. Change its parameter by scrolling along the scrollbars at the top.";
        }

        // Options Menu
        private void optionsToolStripMenuItem1_Click(object sender, EventArgs e)
        {
            Options Form = new Options();
            Form.ShowDialog();
            if (Form.DialogResult == DialogResult.OK)
            {
                output.Write("================= Options =================");
                if (Properties.Settings.Default.calcMethod == 1)
                    output.Write("Area calculating method: Monte Carlo");
                else
                    output.Write("Area calculating method: Riemann");
                output.Write("Amount of dots: " + Properties.Settings.Default.dots);
                output.Write("Show graph-axis: " + Properties.Settings.Default.showAxis);
                output.Write("Anti alias: " + Properties.Settings.Default.antiAlias);
                output.Write("Show grid: " + Properties.Settings.Default.showGrid);
                output.Write("===========================================");

                ReferenceObject.Invalidate();
            }

        }

        // global event if the tool has changed
        private void toolStripButton_Click(object sender, EventArgs e)
        {
            ToolStripButton pressedButton = sender as ToolStripButton;
            if (helpMode)
            {
                output.Write(pressedButton.Tag.ToString());
                return;
            }
            pressedButton.Checked = true;

            foreach (ToolStripItem tsi in toolStrip1.Items)
            {
                ToolStripButton button = tsi as ToolStripButton;
                if (button != null && !tsi.Equals(pressedButton))
                    button.Checked = false;
            }
            embeddedAreas.Clear();
            Tool_InEditMode = false;
            switch (pressedButton.Text)
            {
                case "Edit":
                    Tool_SetToEdit();
                    break;
                case "Sine":
                    Tool_Sine(100, 30, (float)(Math.PI / 50f));
                    break;
                case "Exp1":
                    Tool_Exponential1(200, 1f / 100f);
                    break;
                case "Exp2":
                    Tool_Exponential2(200, 1f / 10000f);
                    break;
                case "Circle":
                    Tool_Circle(200, 100);
                    break;
                case "Linear":
                    Tool_Linear(200, 1);
                    break;
                case "Get Area":
                    Tool_SetToCalcArea();
                    break;
                case "Hysteresis":
                    b.Clear();
                    h.Clear();
                    Tool_SetToHysteresis();
                    break;
            }
            ReferenceObject.Invalidate();
        }

        // Place Func when selected and moving
        private void pictureBox1_MouseMove(object sender, MouseEventArgs e)
        {
            if (selectedFunc != null)
            {
                if (e.X > ReferenceObject.Width - 15 || e.X < 15 || e.Y > ReferenceObject.Height - 15 || e.Y < 15)
                    selectedFunc.MidPos = new PointF(ReferenceObject.Width / 2, ReferenceObject.Height / 2);
                else
                    selectedFunc.MidPos = new PointF(e.X, e.Y);
                pictureBox1.Invalidate();
            }
            MousePos = new PointF(e.X, e.Y);
        }

        // Tool Control
        private void pictureBox1_MouseDown(object sender, MouseEventArgs e)
        {
            switch (e.Button)
            {
                case MouseButtons.Left:
                    switch (curTool)
                    {
                        case Tool.AddFunc:
                            Tool_AddFunc(e);
                            break;
                        case Tool.Edit:
                            Tool_EditFunc(e);
                            break;
                        case Tool.GetArea:
                            Tool_GetArea(e);
                            break;
                    }
                    break;
                case MouseButtons.Right:
                    switch (curTool)
                    {
                        case Tool.AddFunc:
                            break;
                        case Tool.Edit:
                            Tool_DeleteFunc(e);
                            break;
                        case Tool.GetArea:
                            break;
                    }
                    break;
            } 
            pictureBox1.Invalidate();
        }

        // Get the next Func on the drawing area near point
        public int GetNextFunc(List<MathFunc> placedFuncs, PointF point)
        {
            for (int i = 0; i < placedFuncs.Count; i++)
            {
                double dx = placedFuncs[i].MidPos.X - point.X;
                double dy = placedFuncs[i].MidPos.Y - point.Y;
                double distance = Math.Sqrt(dx * dx + dy * dy);
                if (distance < 20)
                    return i;
            }
            return -1;
        }
        public void Tool_AddFunc(MouseEventArgs e)
        {
            output.Write("Function placed at X: " + selectedFunc.MidPos.X + "  Y: " + selectedFunc.MidPos.Y);
            Cursor = Cursors.Default;
            if (placedFuncs.Count > 1)
                placedFuncs.RemoveAt(0);
            selectedFunc.Pen = Pens.Black;
            placedFuncs.Add(selectedFunc);
            if (Tool_InEditMode)
                Tool_SetToEdit();
            else
            {
                MathFunc oldFunc = selectedFunc;
                selectedFunc = oldFunc.RebuildFunc();
                selectedFunc.Invert = oldFunc.Invert;
            }
        }

        public void Tool_EditFunc(MouseEventArgs e)
        {
            int nextFuncIndex = GetNextFunc(placedFuncs, new PointF(e.X, e.Y));
            if (nextFuncIndex >= 0)
            {
                Cursor = Cursors.Hand;
                selectedFunc = placedFuncs[nextFuncIndex];
                selectedFunc.Pen = System.Drawing.Pens.Gray;
                if ((selectedFunc as Sine) != null)
                    Tool_Sine(selectedFunc.XSize, selectedFunc.Multiplier, (selectedFunc as Sine).Omega);
                else if ((selectedFunc as Circle) != null)
                    Tool_Circle(selectedFunc.XSize, selectedFunc.Multiplier);
                else if ((selectedFunc as Linear) != null)
                    Tool_Linear(selectedFunc.XSize, selectedFunc.Multiplier);
                else if (((selectedFunc as Exponential).Exponent % 2) == 0)
                    Tool_Exponential1(selectedFunc.XSize, selectedFunc.Multiplier);
                else
                    Tool_Exponential2(selectedFunc.XSize, selectedFunc.Multiplier);
                selectedFunc.Invert = placedFuncs[nextFuncIndex].Invert;
                placedFuncs.RemoveAt(nextFuncIndex);
                Tool_SetToAdd();
                return;
            }
        }

        public void Tool_DeleteFunc(MouseEventArgs e)
        {
            int nextFuncIndex = GetNextFunc(placedFuncs, new PointF(e.X, e.Y));
            if (nextFuncIndex >= 0)
                placedFuncs.RemoveAt(nextFuncIndex);
        }

        public void Tool_GetArea(MouseEventArgs e)
        {
            if (embeddedAreas.Count > 0)
                for (int i = 0; i < embeddedAreas.Count; i++)
                    if (embeddedAreas[i].IsPointInArea(new PointF(e.X, e.Y)))
                        output.Write("A" + (i + 1) + ": " + Math.Round(embeddedAreas[i].Area, 2));
        }

        // Sets tool to get area mode
        public void Tool_SetToCalcArea()
        {
            if (curTool != Tool.GetArea)
                output.Write("Tool: Get Area");
            curTool = Tool.GetArea;
            sBXAxis.Enabled = false;
            sBOmega.Enabled = false;
            sBMultiplier.Enabled = false;
            selectedFunc = null;
            ReferenceObject.Invalidate();
            embeddedAreas.Clear();
            if (placedFuncs.Count < 2)
            {
                output.Write("There have to be at least two Curves placed");
                return;
            }
            MathFunc func1 = placedFuncs[0];
            MathFunc func2 = placedFuncs[1];
            List<PointF> intersections = GetIntersections(func1, func2);
            if (intersections.Count < 2)
            {
                output.Write("No enclosed area has been found");
                return;
            }
            int percentPerArea = (int)(100f / (float)(intersections.Count - 1));
            for (int i = 0; i < intersections.Count - 1; i++)
            {
                List<PointF> curIntersections = new List<PointF>() { intersections[i], intersections[i + 1] };
                embeddedAreas.Add(new EnclosedArea(placedFuncs, curIntersections));
                toolStripProgressBar.Value += percentPerArea;
            }
            timerProgressbar.Start();
            output.Write("Calculating enclosed area was successful.");
        }

        // Sets tool to edit mode
        public void Tool_SetToEdit()
        {
            if (curTool != Tool.Edit)
                output.Write("Tool: Edit Mode");
            Tool_InEditMode = true;
            curTool = Tool.Edit;
            selectedFunc = null;
            sBMultiplier.Enabled = false;
            sBXAxis.Enabled = false;
            sBOmega.Enabled = false;
            ReferenceObject.Invalidate();
        }

        // Sets tool to function add mode
        public void Tool_SetToAdd()
        {
            if (curTool != Tool.AddFunc)
                output.Write("Tool: Add Function Mode");
            Cursor = Cursors.Hand;
            curTool = Tool.AddFunc;
        }

        // Sine properties
        public void Tool_Sine(int xSize, float multiplier, float omega)
        {
            SetScrollBarMinAndMax(new Range(10, 400), new Range(10, 200));
            SetScrollBarStartValues(xSize, (int)multiplier, (int)((omega * 10000) + 0.1f));
            PointF midPos = new PointF(ReferenceObject.Width / 2, ReferenceObject.Height / 2);
            selectedFunc = new Sine(midPos, xSize, multiplier, omega);
            sBOmega.Enabled = true;
            sBMultiplier.Enabled = true;
            sBXAxis.Enabled = true;
            Tool_SetToAdd();
        }

        // Exponential1 properties
        public void Tool_Exponential1(int xSize, float multiplier)
        {
            SetScrollBarMinAndMax(new Range(0, 200), new Range(66, 200));
            SetScrollBarStartValues(xSize, (int)(1f / multiplier), 0);
            PointF midPos = new PointF(ReferenceObject.Width / 2, ReferenceObject.Height / 2);
            selectedFunc = new Exponential(midPos, xSize, multiplier, 2);
            sBOmega.Enabled = false;
            sBMultiplier.Enabled = true;
            sBXAxis.Enabled = true;
            Tool_SetToAdd();
        }

        // Exponential2 properties
        public void Tool_Exponential2(int xSize, float multiplier)
        {
            SetScrollBarMinAndMax(new Range(0, 200), new Range(5000, 10000));
            SetScrollBarStartValues(xSize, (int)(1f / multiplier), 0);
            PointF midPos = new PointF(ReferenceObject.Width / 2, ReferenceObject.Height / 2);
            selectedFunc = new Exponential(midPos, xSize, multiplier, 3);
            sBOmega.Enabled = false;
            sBMultiplier.Enabled = true;
            sBXAxis.Enabled = true;
            Tool_SetToAdd();
        }

        public void Tool_Circle(int xSize, float multiplier)
        {
            SetScrollBarMinAndMax(new Range(0, 200), new Range(50, 200));
            SetScrollBarStartValues(0, (int)multiplier, 0);
            PointF midPos = new PointF(ReferenceObject.Width / 2, ReferenceObject.Height / 2);
            selectedFunc = new Circle(midPos, xSize, multiplier);
            sBOmega.Enabled = false;
            sBMultiplier.Enabled = true;
            sBXAxis.Enabled = false;
            Tool_SetToAdd();
        }
        
        public void Tool_Linear(int xSize, float multiplier)
        {
            SetScrollBarMinAndMax(new Range(0, 400), new Range(0, 100));
            SetScrollBarStartValues(xSize, (int)multiplier, 0);
            PointF midPos = new PointF(ReferenceObject.Width / 2, ReferenceObject.Height / 2);
            selectedFunc = new Linear(midPos, xSize, multiplier);
            sBOmega.Enabled = false;
            sBMultiplier.Enabled = true;
            sBXAxis.Enabled = true;
            Tool_SetToAdd();
        }

        public void SetScrollBarStartValues(int xAxis, int multiplier, int omega)
        {
            sBXAxis.Value = xAxis;
            sBMultiplier.Value = multiplier;
            if (omega != 0)
                sBOmega.Value = omega;
        }
        public void SetScrollBarMinAndMax(Range xAxis, Range multiplier)
        {
            sBXAxis.Minimum = xAxis.Min;
            sBXAxis.Maximum = xAxis.Max;
            sBMultiplier.Minimum = multiplier.Min;
            sBMultiplier.Maximum = multiplier.Max;
            sBOmega.Minimum = (int)Math.Floor(2 * Math.PI / (double)xAxis.Max * 10000);
            sBOmega.Maximum = (int)Math.Floor(2 * Math.PI / ((double)xAxis.Max / 4) * 10000);
        }
        // Paint event
        private void pictureBox1_Paint(object sender, PaintEventArgs e)
        {

            Graphics g = e.Graphics;
            if (Properties.Settings.Default.antiAlias)
                g.SmoothingMode = System.Drawing.Drawing2D.SmoothingMode.AntiAlias;
            if (Properties.Settings.Default.showGrid)
            {                
                grid.DrawGridBySize(Properties.Settings.Default.gridGapSize, g);
            }
            for (int i = 0; i < embeddedAreas.Count; i++)
            {
                embeddedAreas[i].DrawInformation(g, i + 1);
            }
            foreach (MathFunc func in placedFuncs)
            {
                func.DrawFunc(g);
                if (curTool == Tool.Edit)
                    g.DrawEllipse(Pens.MidnightBlue, func.MidPos.X - 20, func.MidPos.Y - 20, 40, 40);
            }
            if (selectedFunc != null)
                selectedFunc.DrawFunc(g);
            g.DrawString("X: " + MousePos.X + "  Y: " + MousePos.Y, new Font("", 8, FontStyle.Bold), Brushes.Black, new PointF(5, 5));
            //hysterese
            if (curTool == Tool.Hysteresis)
                drawHysterese = hysterese.DrawHysterese(g, b, h, Properties.Settings.Default.showAxis);
            
        }

        // Scrollbar multiplier
        private void sBMultiplier_Scroll(object sender, ScrollEventArgs e)
        {
            float multiplier = sBMultiplier.Value;
            if (selectedFunc as Exponential != null)
                multiplier = 1f / multiplier;
            if (selectedFunc as Circle != null)
                selectedFunc.XSize = 2 * (int)multiplier;
            if (selectedFunc as Linear != null)
                multiplier = multiplier / 50;
            selectedFunc.Multiplier = multiplier;
            lMultiplier.Text = multiplier.ToString();
            ReferenceObject.Invalidate();
        }
        
        // Scrollbar range of x axis
        private void sBXAxis_Scroll(object sender, ScrollEventArgs e)
        {
            selectedFunc.XSize = sBXAxis.Value;
            lXAxis.Text = sBXAxis.Value.ToString();
            ReferenceObject.Invalidate();
        }

        // Scrollbar omega
        private void sBOmega_Scroll(object sender, ScrollEventArgs e)
        {
            (selectedFunc as Sine).Omega = sBOmega.Value / 10000f;
            lOmega.Text = sBOmega.Value.ToString();
            ReferenceObject.Invalidate();
        }

        // Invert button
        private void btnInvert_Click(object sender, EventArgs e)
        {
            if (selectedFunc != null)
            {
                selectedFunc.Invert *= -1;
                ReferenceObject.Invalidate();
            }
        }

        // Returns the intersections of two functions
        List<PointF> GetIntersections(MathFunc func1, MathFunc func2)
        {
            int xMin = -func1.XSize / 2;
            int xMax = func1.XSize / 2;
            PointF p1 = func1.MidPos;
            PointF p2 = func2.MidPos;
            bool firstHigher = false;
            bool funcsOverlapping = false;
            List<PointF> allPoints = new List<PointF>();
            List<PointF> intersections = new List<PointF>();
            int index = 0;
            for (float x = xMin; x < xMax; x += 0.1f)
            {   
                float y1 = func1.GetY(x);
                float x2 = (float)(p1.X + x - p2.X);
                float y2 = func2.GetY(x2) + p2.Y - p1.Y;
                y1 += func1.MidPos.Y;
                y2 += func1.MidPos.Y;
                if (x2 < (func2.XSize / 2) && x2 > (-func2.XSize / 2))
                {
                    if (!funcsOverlapping)
                    {
                        funcsOverlapping = true;
                        firstHigher = (y1 < y2) ? true : false;
                    }
                    else if (y1 > y2 && firstHigher)
                    {
                        intersections.Add(allPoints[index - 1]);
                        firstHigher = false;
                    }
                    else if (y1 < y2 && !firstHigher)
                    {
                        intersections.Add(allPoints[index - 1]);
                        firstHigher = true;
                    }
                }
                allPoints.Add(new PointF(x + func1.MidPos.X, (y1 + y2) / 2));
                index++;
                
            }
            return intersections;
        }
        

        // Some Events
        private void saveOutputToolStripMenuItem_Click(object sender, EventArgs e)
        {
            output.SaveOutputAsTxt();
        }
        private void loadOutputToolStripMenuItem_Click(object sender, EventArgs e)
        {
            output.Opentxt();
        }
  
        private void fileLoadToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Tool_SetToHysteresis();
        }

        private void Form1_Resize(object sender, EventArgs e)
        {
            pictureBox1.Invalidate();
        }

        public void Tool_SetToHysteresis()
        {
            if (curTool != Tool.Hysteresis)
                output.Write("Tool: Load Hysteresis");
            curTool = Tool.Hysteresis;
            sBXAxis.Enabled = false;
            sBOmega.Enabled = false;
            sBMultiplier.Enabled = false;
            selectedFunc = null;
            if (hysterese.Readfile(out b, out h))

            {
                output.Write("File was successfully loaded");
                placedFuncs.Clear();
                ReferenceObject.Invalidate();
                double foo = hysterese.CalcArea(b, h);
                output.Write("A: "+Math.Round(foo,2).ToString()+" mJ / m^3");
            }     
        }

        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            AboutBox1 about = new AboutBox1();
            about.ShowDialog();
        }

        private void timerProgressbar_Tick(object sender, EventArgs e)
        {
            toolStripProgressBar.Value = 0;
            timerProgressbar.Stop();
        }

        private void helpToolStripMenuItem_Click(object sender, EventArgs e)
        {
            helpMode = !helpMode;
            helpToolStripMenuItem.Checked = helpMode;
            if (helpMode)
            {
                curTool = Tool.None;
                selectedFunc = null;
                ReferenceObject.Invalidate();
                output.Write("HelpMode on");
                output.Write("================= Help Mode =================");
                output.Write("Help Mode is currently active.");
                output.Write("To switch it off click Help Mode in the '?' tab again.");
                output.Write("Click on a tool at the right side to get infos");
                output.Write("while help mode is active.");
                output.Write("=============================================");
            }
            else
                output.Write("HelpMode off");
        }

        private void saveFileAsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            output.SaveAsPicture();
        }

        private void showExampleToolStripMenuItem_Click(object sender, EventArgs e)
        {
            curTool = Tool.GetArea;
            placedFuncs.Clear();
            embeddedAreas.Clear();
            output.Write("Real Area: PI * 100² / 2 = " + (Math.PI * 100 * 100 / 2));
            placedFuncs.Add(new Circle(new PointF(pictureBox1.Width / 2, pictureBox1.Height / 2), 200, 100));
            placedFuncs.Add(new Linear(new PointF(pictureBox1.Width / 2, pictureBox1.Height / 2), 300, 0));
            embeddedAreas.Add(new EnclosedArea(placedFuncs, new List<PointF>() { new PointF(pictureBox1.Width / 2 - 100, pictureBox1.Height / 2), new PointF(pictureBox1.Width / 2 + 100, pictureBox1.Height / 2) }));
            pictureBox1.Invalidate();
        }

    }
}